/**
 * Copyright (c) 2008 Washington University
 */
package org.nrg.xnat.filestore;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.URI;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.Collection;
import java.util.Map;

import javax.activation.DataSource;

import org.nrg.xnat.tags.TagLabel;

/**
 * @author Kevin A. Archie <karchie@npg.wustl.edu>
 */
public interface FileStore {
    /**
     * Stores the given resource
     * @param uri URI to use for the resource
     * @param file the underlying resource
     * @param mediaType MIME media type for the resource (null for default "application/octet-stream")
     * @throws SQLException
     * @throws ForbiddenException if the URI is already registered
     */
    public void create(Connection c, URI uri, File file, String mediaType) throws SQLException,ForbiddenException;
    
    /**
     * Stores the given resource with default MIME media type
     * @param uri URI to use for the resource
     * @param file the underlying resource
     * @throws SQLException
     * @throws ForbiddenException if the URI is already registered
     */
    public void create(Connection c, URI uri, File file) throws SQLException,ForbiddenException;
    
    /**
     * Stores the given byte stream as a new resource
     * @param uri URI by which this resource will be exported
     * @param ds DataSource providing the resource data
     * @throws SQLException
     * @throws IOException
     * @throws ForbiddenException if the URI is already assigned to some resource
     */
    public void create(Connection c, URI uri, DataSource ds) throws SQLException,IOException,ForbiddenException;

    /**
     * Replaces an existing resource
     * @param uri URI for the resource to be replaced
     * @param ds DataSource providing the new resource data
     * @return true if this is a new resource, false if the resource already existed
     * @throws SQLException
     * @throws IOException
     * @throws ForbiddenException
     */
    public boolean update(Connection c, URI uri, DataSource ds) throws SQLException,IOException,ForbiddenException;
    
    /**
     * Gets the data for the named resource
     * @param uri URI for the resource
     * @return the data file
     * @throws SQLException
     * @throws IOException
     */
    public File getFile(Connection c, URI uri) throws SQLException,FileNotFoundException;
    
    /**
     * Retrieves an Activation DataSource for the named resource
     * @param uri exported URI for the resource
     * @return DataSource for the resource
     * @throws SQLException
     * @throws FileNotFoundException if the resource is not available
     */
    public DataSource getDataSource(Connection c, final URI uri) throws SQLException,FileNotFoundException;
    
    /**
     * Removes the named resources and all their metadata from the database.
     * @param removeFile true if the file should be removed from the filesystem.
     * @param uris resources to be removed
     * @throws SQLException
     */
    public void delete(Connection c, boolean removeFile, URI...uris) throws SQLException,FileNotFoundException;
    
    /**
     * Removes the named resources, all their metadata, and the underlying resources.
     * @param removeFile true if the file should be removed from the filesystem.
     * @param uris resources to be removed
     * @throws SQLException
     */
    public void delete(Connection c, boolean removeFile, Collection<URI> uris) throws SQLException,FileNotFoundException;
    
    /**
     * Retrieves the metadata (tags) for the given object
     * @param uri resource identifier
     * @return a Map TagLabel->Collection<String> of the metadata for the named object
     * @throws SQLException
     * @throws FileNotFoundException if the resource is not registered
     */
    public Map<TagLabel,Collection<String>> getMetadata(Connection c, URI uri) throws SQLException,FileNotFoundException;
    
    /**
     * Applies the given metadata operations to the database
     * @param uri resource for which metadata should be modified
     * @param ops changes to make to the metadata
     * @throws SQLException
     * @throws FileNotFoundException if the named resource is not in the database
     */
    public void updateMetadata(Connection c, URI uri, MetadataOp...ops) throws SQLException,FileNotFoundException;
}
