/**
 * Copyright (c) 2008 Washington University
 */
package org.nrg.xnat.filestore;

import java.io.ByteArrayOutputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Collection;
import java.util.Iterator;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.activation.DataHandler;
import javax.mail.BodyPart;
import javax.mail.MessagingException;
import javax.mail.internet.InternetHeaders;
import javax.mail.internet.MimeBodyPart;
import javax.mail.internet.MimeMultipart;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;
import javax.xml.XMLConstants;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.transform.Source;
import javax.xml.transform.stream.StreamSource;
import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;

import org.apache.commons.lang.RandomStringUtils;
import org.apache.log4j.Logger;
import org.dom4j.Attribute;
import org.dom4j.Document;
import org.dom4j.DocumentException;
import org.dom4j.DocumentFactory;
import org.dom4j.Element;
import org.dom4j.Namespace;
import org.dom4j.QName;
import org.dom4j.io.OutputFormat;
import org.dom4j.io.SAXReader;
import org.dom4j.io.XMLWriter;
import org.nrg.xnat.search.Search;
import org.nrg.xnat.tags.InvalidTagLabelException;
import org.nrg.xnat.tags.TagLabel;
import org.nrg.xnat.tags.TagLabelException;
import org.nrg.xnat.tags.TagLabelsServlet;
import org.xml.sax.SAXException;


/**
 * HttpServlet that handles file, bundle and metadata operations:
 * upload (POST)
 * download (GET)
 * modify (PUT)
 * delete (DELETE) 
 * @author Kevin A. Archie <karchie@npg.wustl.edu>
 *
 */
public final class FileStoreServlet extends HttpServlet {
    private static final long serialVersionUID = 1L;

    private static final String JNDI_JDBC_DATASOURCE = "java:comp/env/jdbc/datasource";
    private static final String JNDI_XNAT_STORE = "java:comp/env/xnat/store";

    private static final String HTTP_LOCATION = "Location";
    private static final String MIME_CONTENT_DISPOSITION = "Content-Disposition";
    private static final String MIME_CONTENT_TYPE = "Content-Type";
    private static final String CONTENT_TYPE_METADATA = "application/x-xnat-metadata+xml";
    private static final String CONTENT_TYPE_DATA_DEFAULT = "application/octet-stream";
    private static final String CONTENT_DISPOSITION_DATA = "inline";
    private static final String CONTENT_DISPOSITION_METADATA = "x-xnat-metadata";

    // Here's some serious Perl-style regexp line noise.  In pieces:
    // .*;\\s*  a non-capturing group that indicates that anything may precede the filename
    //          spec, so long as it ends in ; plus some optional whitespace;
    //
    // filename\\s*\\\s*  next is filename = , optionally with whitespace around the =
    //
    // \"([^\\\"]*)\"   next is a capturing group for the filename text (surrounded by quotes)
    //                  the filename can contain any character other than ", though some other
    //                  characters (/, \) will probably cause trouble down the line
    // 
    // (?:;|\\s.*)?  optional, noncapturing group for whitespace or ; (parameter separator)
    private static final Pattern DISPOSITION_FILENAME_PATTERN = Pattern.compile(".*;\\s*filename\\s*=\\s*\"([^\\\"]*)\"(?:;|\\s.*)?");

    private static final String XSD_METADATA = "xnat-metadata.xsd";
    private static final SAXParserFactory metadataParserFactory;
    static {
	metadataParserFactory = SAXParserFactory.newInstance();

	final ClassLoader cl = FileStoreServlet.class.getClassLoader();
	final InputStream is = cl.getResourceAsStream(XSD_METADATA);
	if (null != is) {
	    final Source source = new StreamSource(is);
	    final SchemaFactory factory = SchemaFactory.newInstance(XMLConstants.W3C_XML_SCHEMA_NS_URI);
	    final Schema schema;
	    try {
		schema = factory.newSchema(source);
		metadataParserFactory.setSchema(schema);
		metadataParserFactory.setNamespaceAware(true);
		metadataParserFactory.setValidating(true);
	    } catch (SAXException e) {
		Logger.getLogger(FileStoreServlet.class).error("Unable to load schema " + XSD_METADATA, e);
	    }
	}
	try { is.close(); } catch (IOException ignore) {}
    }

    private static final Namespace NAMESPACE_XE = new Namespace(null, "http://nrg.wustl.edu/xe");
    private static final QName ROOT_METADATA = new QName("Metadata", NAMESPACE_XE);
    private static final QName ELEM_METADATA = new QName("Tag", NAMESPACE_XE);
    private static final String ATTR_METADATA_LABEL = "Label";
    private static final QName ELEM_METADATA_VALUE = new QName("Value", NAMESPACE_XE);
    private static final QName ELEM_METADATA_VALUE_OP_REPLACE = new QName("Replace", NAMESPACE_XE);
    private static final QName ELEM_METADATA_VALUE_OP_DELETE = new QName("Delete", NAMESPACE_XE);

    private static final String GET_PARAM_PART = "part";
    private static final String GET_PARAM_PART_METADATA = "metadata";
    private static final String GET_PARAM_PART_RESOURCE = "resource";

    private static final String SQL_GET_LABEL_ID = "SELECT tag_label_id FROM tag_labels WHERE tag_label_lc = LOWER(?)";

    private final Logger log = Logger.getLogger(FileStoreServlet.class);
    private FileStore store = null;
    private DataSource ds = null;

    public void init(final ServletConfig config) throws ServletException {
	super.init(config);
	try {
	    final InitialContext ic = new InitialContext();
	    ds = (DataSource)ic.lookup(JNDI_JDBC_DATASOURCE);
	    store = (FileStore)ic.lookup(JNDI_XNAT_STORE);
	} catch (NamingException e) {
	    throw new ServletException(e);
	}
    }


    private void storeData(final Connection dbc, final HttpServletRequest req, final HttpServletResponse res, final URI uri, final boolean knownNew)
    throws IOException {
	final String contentType = getContentType(req);
	boolean isNewResource = knownNew;
	try {
	    if (contentType.startsWith("multipart/")) {
		if (true) {
		    log.error("unable to process request of type " + contentType);
		    res.sendError(HttpServletResponse.SC_NOT_IMPLEMENTED, "MIME multipart messages not supported");
		    return;
		}
		final MimeMultipart mmp = new MimeMultipart(new InputStreamDataSource(req.getInputStream(), contentType));

		final int nParts = mmp.getCount();
		if (nParts < 1 || 2 < nParts) {
		    log.info("received multipart message with strange part count: " + nParts);
		    res.sendError(HttpServletResponse.SC_BAD_REQUEST, "multipart message must have 1 or 2 parts");
		}

		log.debug("decoding multipart request with " + nParts + " parts");

		BodyPart dataPart = null;
		MetadataOp[] metadata = null;


		for (int i = 0; i < nParts; i++) {
		    final BodyPart part = mmp.getBodyPart(i);
		    final String partType = getContentType(part);
		    final String disposition = part.getDisposition();
		    if (partType.startsWith(CONTENT_TYPE_METADATA)) {
			if (null != metadata) {
			    final String message = "Multiple metadata parts defined in message entity";
			    log.debug(message);
			    res.sendError(HttpServletResponse.SC_BAD_REQUEST, message);
			    return;
			}
			try {
			    metadata = parseMetadataOps(dbc, part.getInputStream());
			} catch (SAXException e) {
			    log.debug(e);
			    res.sendError(HttpServletResponse.SC_BAD_REQUEST, e.getMessage());
			    return;
			}
		    } else if (disposition == null || disposition.startsWith(CONTENT_DISPOSITION_DATA)) {
			if (null != dataPart) {
			    final String message = "Multiple data parts defined in message entity";
			    log.debug(message);
			    res.sendError(HttpServletResponse.SC_BAD_REQUEST, message);
			    return;
			}
			dataPart = part;
		    } else {
			final String message = "Unsupported Content-Type,Content-Disposition pair: " + partType + ", " + disposition;
			log.info(message);
			res.sendError(HttpServletResponse.SC_BAD_REQUEST, message);
			return;
		    }
		}
		if (null == dataPart && null == metadata) {
		    log.info("empty multipart message");
		    res.sendError(HttpServletResponse.SC_BAD_REQUEST, "empty multipart message");
		    return;
		}
		isNewResource |= store.update(dbc, uri, null == dataPart ? null : dataPart.getDataHandler().getDataSource());
		if (null != metadata) {
		    try {
			store.updateMetadata(dbc, uri, metadata);
		    } catch (FileNotFoundException e) {
			log.error("database does not contain just-registered resource ", e);
			res.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, e.getMessage());
			return;
		    }
		}
	    } else {
		// Message entity is a single part.  Store the request (metadata-only, or data-only).
		if (contentType.startsWith(CONTENT_TYPE_METADATA)) {
		    final MetadataOp[] metadata;
		    try {
			metadata = parseMetadataOps(dbc, req.getInputStream());
		    } catch (SAXException e) {
			log.error(e);
			res.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, e.getMessage());
			return;
		    }
		    try {
			isNewResource |= store.update(dbc, uri, null);
			store.updateMetadata(dbc, uri, metadata);
		    } catch (FileNotFoundException e) {
			log.debug(e);
			res.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, "resource not registered in database: " + uri);
			return;
		    }
		} else {
		    isNewResource |= store.update(dbc, uri, new InputStreamDataSource(req.getInputStream(), contentType));
		}
		if (isNewResource) {
		    // If we created a new resource, should return 201 Created, and put
		    // the URI in the response entity and the Location: header.
		    res.setStatus(HttpServletResponse.SC_CREATED);

		    // Constructing an external URI is annoying.
		    final URL reqURL = new URL(req.getRequestURL().toString());
		    final String protocol = reqURL.getProtocol();
		    final String host = reqURL.getHost();
		    final int port = reqURL.getPort();

		    final URL location = new URL(protocol, host, port, req.getServletPath() + "/" + uri.toString());
		    res.setHeader(HTTP_LOCATION, location.toString());
		    res.getOutputStream().write(location.toString().getBytes());	    
		} else {
		    res.setStatus(HttpServletResponse.SC_NO_CONTENT);
		}
	    }

	    doDBCheckpoint(dbc);
	} catch (DocumentException e) {		// error in received XML document
	    log.debug(e);
	    res.sendError(HttpServletResponse.SC_BAD_REQUEST, e.getMessage());
	    return;
	} catch (TagLabelException e) {	// passed to this level, always indicates bad tag label
	    log.debug(e);
	    res.sendError(HttpServletResponse.SC_BAD_REQUEST, e.getMessage());
	    return;
	} catch (MessagingException e) {	// always indicates error in MIME encapsulation
	    log.debug(e);
	    res.sendError(HttpServletResponse.SC_BAD_REQUEST, e.getMessage());
	    return;
	} catch (ForbiddenException e) {	// always indicates failed access to a resource
	    log.debug(e);
	    res.sendError(HttpServletResponse.SC_FORBIDDEN, e.getMessage());
	    return;
	} catch (SQLException e) {		// always indicates error accessing database
	    log.error(e);
	    res.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, e.getMessage());
	    return;
	}
    }


    private void doDBCheckpoint(final Connection c) {
	try {
	    final Statement s = c.createStatement();
	    try {
		s.execute("CHECKPOINT;");
	    } catch (SQLException e) {
		Logger.getLogger(TagLabelsServlet.class).warn(e);
	    } finally {
		try { s.close(); } catch (SQLException ignore) {}
	    }
	} catch (SQLException e) {
	    Logger.getLogger(TagLabelsServlet.class).warn("CHECKPOINT failed", e);
	}
    }


    /**
     * POST /<servlet-context-path>
     * Defines a resource for storage in the repository, assigning a new URI.
     * Client may supply data and/or metadata
     */
    protected void doPost(final HttpServletRequest req, final HttpServletResponse res) throws IOException {
	// Allow POST only to servlet context root
	if (!Pattern.matches(req.getServletPath() + "/?", req.getRequestURI())) {
	    res.sendError(HttpServletResponse.SC_METHOD_NOT_ALLOWED, "POST allowed only to " + req.getServletPath());
	    return;
	}

	// TODO: if multipart, look through parts for disposition too
	final String contentDisposition = req.getHeader(MIME_CONTENT_DISPOSITION);

	final String filename;
	if (null != contentDisposition) {
	    final Matcher matcher = DISPOSITION_FILENAME_PATTERN.matcher(contentDisposition);
	    filename = matcher.matches() ? matcher.group(1) : null;
	} else {
	    filename = null;
	}

	final Connection c;
	try {
	    c = ds.getConnection();
	} catch (SQLException e) {
	    log.error("Unable to make database connection", e);
	    res.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, e.getMessage());
	    return;
	}

	try {
	    final URI uri = makeURI(c, filename, null != req.getParameter("obscure"));
	    storeData(c, req, res, uri, true);
	} catch (SQLException e) {
	    log.error(e);
	    res.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, e.getMessage());
	    return;
	} catch (URISyntaxException e) {
	    log.error(e);
	    res.sendError(HttpServletResponse.SC_BAD_REQUEST, "Invalid proposed filename " + filename);
	    return;
	} finally {
	    try { c.close(); } catch (SQLException ignore) {}
	}
    }


    /**
     * PUT resource-uri
     * Stores or replaces data and/or metadata for a named resource
     */
    protected void doPut(final HttpServletRequest req, final HttpServletResponse res) throws IOException {
	final URI uri;
	try {
	    uri = getLocalURI(req);
	} catch (URISyntaxException e) {
	    log.error(e);
	    res.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, e.getMessage());
	    return;
	}

	try {
	    final Connection c = ds.getConnection();
	    try {
		storeData(c, req, res, uri, false);
	    } finally {
		try { c.close(); } catch (SQLException ignore) {}
	    }
	} catch (SQLException e) {
	    log.error("database error on PUT", e);
	    res.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, e.getMessage());
	    return;
	}
    }


    private final class InputStreamDataSource implements javax.activation.DataSource {
	private final String name, contentType;
	private final InputStream is;

	InputStreamDataSource(final InputStream is, final String contentType, final String name) {
	    this.contentType = contentType;
	    this.is = is;
	    this.name = name;
	}

	InputStreamDataSource(final InputStream is, final String contentType) {
	    this(is, contentType, null);
	}

	final public String getContentType() { return contentType; }
	final public String getName() { return name; }
	final public InputStream getInputStream() { return is; }
	final public OutputStream getOutputStream() { throw new UnsupportedOperationException(); }
    }


    /**
     * This request couldn't be satisifed as a normal GET, because no real resource exists with that URI.
     * Look for partial path completions instead.
     * @param path URI of resource
     * @param req received HTTP request
     * @param res HTTP response
     * @throws FileNotFoundException if we can't find any partial path matches
     */
    private final void doPartialPathGet(final Connection c, final URI uri, final HttpServletRequest req, final HttpServletResponse res)
    throws FileNotFoundException,SQLException {
	final Iterator<URI> completionsi = Search.getCompletions(c, uri).resourceIterator();

	if (!completionsi.hasNext()) {
	    throw new FileNotFoundException(req.getRequestURI());
	}

	// TODO: sort the results
	if (null == req.getParameter("recursive")) {
	    throw new UnsupportedOperationException("partial path completion not implemented");
	} else {
	    throw new UnsupportedOperationException("recursive partial path completion not implemented");
	}
    }


    /**
     * GET resource-uri
     * Retrieves a resource (data file) from the repository, optionally with associated metadata
     */
    protected void doGet(final HttpServletRequest req, final HttpServletResponse res) throws IOException {
	final URI uri;
	try {
	    uri = getLocalURI(req);
	} catch (URISyntaxException e) {
	    log.error("Bad request URI", e);
	    res.sendError(HttpServletResponse.SC_BAD_REQUEST, "Bad request URI " + e.getInput() + ": " + e.getMessage());
	    return;
	}
	final Connection c;
	try {
	    c = ds.getConnection();
	} catch (SQLException e) {
	    log.error(e);
	    res.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, e.getMessage());
	    return;
	}

	try {
	    int nParts = 0;
	    final Document metadata;
	    final javax.activation.DataSource data;
	    final Map<?,?> params = req.getParameterMap();

	    final Collection<String> parts;
	    if (params.containsKey(GET_PARAM_PART)) {
		parts = Arrays.asList((String[])params.get(GET_PARAM_PART));
	    } else {
		parts = new ArrayList<String>(0);
	    }

	    // Send data by default (no part= parameters)
	    // or if user asked explicitly for it (part=resource)
	    if (parts.isEmpty() || parts.contains(GET_PARAM_PART_RESOURCE)) {
		try {
		    data = store.getDataSource(c, uri);
		    nParts++;
		} catch (FileNotFoundException e) {
		    // Maybe this is a partial path we can complete.
		    doPartialPathGet(c, uri, req, res);
		    return;
		}
	    } else {
		data = null;
	    }

	    if (parts.contains(GET_PARAM_PART_METADATA)) {
		try {
		    metadata = getMetadataDocument(c, uri);
		} catch (FileNotFoundException e) {
		    log.debug("Requested metadata for unknown resource " + uri);
		    throw e;
		} catch (TagLabelException e) {
		    log.error(e);
		    res.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, e.getMessage());
		    return;
		}
		nParts++;
	    } else {
		metadata = null;
	    }

	    final MimeMultipart multipart;
	    if (nParts > 1) {
		multipart = new MimeMultipart("related");
	    } else {
		multipart = null;
	    }

	    // Package data/bundle and/or metadata:
	    // if one piece, can send as simple entity
	    // otherwise, must make mime multipart
	    if (null != data) {
		// Data object download
		final DataHandler dh = new DataHandler(data);
		if (null == multipart) {
		    // message body is encoded data
		    res.setContentType(data.getContentType());
		    res.setHeader(MIME_CONTENT_DISPOSITION, CONTENT_DISPOSITION_DATA);
		    try {
			dh.writeTo(res.getOutputStream());
		    } catch (IOException e) {
			log.error(e);
			res.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, e.getMessage());
			return;
		    }
		} else {
		    // Build a BodyPart and add it to the multipart
		    final MimeBodyPart mbp = new MimeBodyPart();
		    mbp.setDataHandler(dh);
		    mbp.setHeader(MIME_CONTENT_TYPE, dh.getContentType());
		    mbp.setHeader(MIME_CONTENT_DISPOSITION, CONTENT_DISPOSITION_DATA);
		    multipart.addBodyPart(mbp);
		}
	    }

	    if (null != metadata) {
		// metadata document
		if (null == multipart) {
		    res.setContentType(CONTENT_TYPE_METADATA);
		    res.setHeader(MIME_CONTENT_DISPOSITION, CONTENT_DISPOSITION_METADATA);
		    try {
			log.debug("embedding metadata XML directly in response entity");
			metadata.write(res.getWriter());
		    } catch (IOException e) {
			log.error(e);
			res.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, e.getMessage());
			return;
		    }
		} else {
		    try {
			multipart.addBodyPart(metadataBodyPart(metadata));
		    } catch (IOException e) {
			log.error(e);
			res.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, e.getMessage());
			return;
		    }
		}
	    }

	    if (null != multipart) {
		res.setContentType(multipart.getContentType());
		multipart.writeTo(res.getOutputStream());
	    }
	} catch (FileNotFoundException e) {
	    res.sendError(HttpServletResponse.SC_NOT_FOUND, uri.toString());
	    return;
	} catch (MessagingException e) {
	    log.error(e);
	    res.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, e.getMessage());
	    return;    
	} catch (SQLException e) {
	    log.error(e);
	    res.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, e.getMessage());
	    return;
	} finally {
	    try { c.close(); } catch (SQLException e) { log.debug(e); }
	}
    }


    /**
     * DELETE resource-uri
     * Removes a resource and its associated metadata from the repository
     */
    protected void doDelete(final HttpServletRequest req, final HttpServletResponse res) throws IOException {
	try {
	    final URI uri = getLocalURI(req);
	    final Connection c = ds.getConnection();
	    try {
		store.delete(c, true, uri);
		doDBCheckpoint(c);
	    } finally {
		c.close();
	    }
	} catch (FileNotFoundException e) {
	    log.debug(e);
	    res.sendError(HttpServletResponse.SC_NOT_FOUND, e.getMessage());
	    return;
	} catch (URISyntaxException e) {
	    log.debug(e);
	    res.sendError(HttpServletResponse.SC_BAD_REQUEST, e.getMessage());
	    return;
	} catch (SQLException e) {
	    log.error(e);
	    res.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, e.getMessage());
	    return;
	}
    }


    private Document getMetadataDocument(final Connection c, final URI uri) throws SQLException,FileNotFoundException,TagLabelException {
	final Map<TagLabel,Collection<String>> m = store.getMetadata(c, uri);

	final Document d = DocumentFactory.getInstance().createDocument();
	final Element root = d.addElement(ROOT_METADATA);
	for (final Map.Entry<TagLabel,Collection<String>> me : m.entrySet()) {
	    final Element tag = root.addElement(ELEM_METADATA);
	    tag.addAttribute(ATTR_METADATA_LABEL, me.getKey().toString());
	    for (final String v : me.getValue()) {
		final Element value = tag.addElement(ELEM_METADATA_VALUE);
		value.setText(v);
	    }
	}

	return d;
    }


    private Document readMetadataDocument(final InputStream is) throws SAXException,DocumentException {
	final SAXParser parser;
	try {
	    parser = metadataParserFactory.newSAXParser();
	} catch (ParserConfigurationException e) {
	    throw new SAXException(e);
	}

	final SAXReader saxReader = new SAXReader(parser.getXMLReader());
	log.debug("Using " + saxReader.getClass().getName() + "/" + saxReader.getXMLReader().getClass().getName()
		+ " to validate against schema " + parser.getSchema());

	return saxReader.read(is);
    }


    private MetadataOp[] parseMetadataOps(final Connection c, final InputStream is) throws SQLException,SAXException,DocumentException,TagLabelException {
	final Document doc = readMetadataDocument(is);
	final Element root = doc.getRootElement();

	final Collection<MetadataOp> ops = new ArrayList<MetadataOp>();

	final PreparedStatement getLabelID = c.prepareStatement(SQL_GET_LABEL_ID);
	try {
	    for (final Iterator<?> tagi = root.elementIterator(); tagi.hasNext(); ) {
		final Element labele = (Element)tagi.next();

		final Attribute labela = labele.attribute(ATTR_METADATA_LABEL);
		assert null != labela;	// would be caught in validation
		final TagLabel label = new TagLabel(labela.getText());

		// get the label ID for the given tag label
		final int labelID;
		getLabelID.setString(1, label.toString());
		final ResultSet lidrs = getLabelID.executeQuery();
		try {
		    if (!lidrs.next()) {
			throw new InvalidTagLabelException(label.toString());
		    }
		    labelID = lidrs.getInt(1);
		} finally {
		    try { lidrs.close(); } catch (SQLException ignore) {}
		}

		for (final Iterator<?> vali = labele.elementIterator(); vali.hasNext(); ) {
		    final Element vale = (Element)vali.next();
		    if (null != vale.attribute(ELEM_METADATA_VALUE_OP_REPLACE)) {
			// TODO: verify that "change" attribute is null
			throw new UnsupportedOperationException("REPLACE metadata operation not yet implemented");
		    } else if (null != vale.attribute(ELEM_METADATA_VALUE_OP_DELETE)) {
			// "delete" must be null
			throw new UnsupportedOperationException("DELETE metadata operation not yet implemented");
		    } else {
			ops.add(new MetadataAssignOp(c, labelID, vale.getText()));
		    }
		}
	    }
	} finally {
	    try { getLabelID.close(); } catch (SQLException ignore) {}
	}

	return ops.toArray(new MetadataOp[0]);
    }


    private final MimeBodyPart metadataBodyPart(final Document d)
    throws MessagingException,IOException {
	final InternetHeaders headers = new InternetHeaders();
	headers.addHeader(MIME_CONTENT_TYPE, CONTENT_TYPE_METADATA);
	headers.addHeader(MIME_CONTENT_DISPOSITION, CONTENT_DISPOSITION_METADATA);
	final ByteArrayOutputStream bos = new ByteArrayOutputStream();
	final OutputFormat of = OutputFormat.createPrettyPrint();
	of.setSuppressDeclaration(false);
	final XMLWriter writer = new XMLWriter(bos, of);
	writer.write(d);
	return new MimeBodyPart(headers,bos.toByteArray());
    }


    private final static URI getLocalURI(final HttpServletRequest req) throws URISyntaxException {
	final String servletPath = req.getServletPath() + "/";
	final String suri = req.getRequestURI().replaceFirst(servletPath, "");
	if (!Utils.isValidPath(suri)) {
	    throw new URISyntaxException(suri, "not a valid file store URI");
	}
	return new URI(suri);
    }


//  null Content-Type should always be mapped to the default type, application/octet-stream.
    private final static String getContentType(final String s) {
	return null == s ? CONTENT_TYPE_DATA_DEFAULT : s;
    }

    private final static String getContentType(final HttpServletRequest req) {
	return getContentType(req.getContentType());
    }

    private final static String getContentType(final BodyPart part) {
	try {
	    return getContentType(part.getContentType());
	} catch (MessagingException e) {
	    Logger.getLogger(FileStoreServlet.class).debug(e);
	    return CONTENT_TYPE_DATA_DEFAULT;
	}
    }

    // Use compact ISO-8601 date/time stamp
    private static final DateFormat DATETIME_FORMAT = new SimpleDateFormat("yyyyMMdd'T'HHmmss'Z'/");

    private final URI makeURI(final Connection c, final String filename, final boolean obscure) throws SQLException,URISyntaxException {
	final StringBuilder sb = new StringBuilder();
	if (obscure) {
	    // You asked for it...
	    sb.append(RandomStringUtils.randomAlphanumeric(16));
	} else {
	    sb.append(DATETIME_FORMAT.format(Calendar.getInstance().getTime()));
	    if (null == filename) {
		sb.append(RandomStringUtils.randomAlphanumeric(8));
	    } else {
		if (Utils.isValidPath(filename)) {
		    sb.append(filename);
		} else {
		    throw new URISyntaxException(filename, "not acceptable file store pathname");
		}
	    }
	}


	// Now skip ahead until we find a safe resource
	try {
	    final String base = sb.toString();
	    URI uri = new URI(base);
	    for (int i = 1; ; i++) {
		try {
		    store.create(c, uri, (javax.activation.DataSource)null);
		    return uri;
		} catch (ForbiddenException uriIsTaken) {
		} catch (IOException e) {
		    log.warn(e);
		}
		uri = new URI(String.format("%s_%d", base, i));
	    }
	} catch (URISyntaxException e) {
	    throw new RuntimeException(e);	// TODO: this shouldn't happen
	}
    }
}
