/**
 * $Id$
 * Copyright (c) 2008 Washington University
 */
package org.nrg.xnat.filestore;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Collection;
import java.util.HashSet;
import java.util.regex.Pattern;

import javax.activation.DataSource;


import org.apache.commons.io.IOUtils;
import org.apache.log4j.Logger;


/**
 * @author Kevin A. Archie <karchie@npg.wustl.edu>
 */
public class URIPathFileStore extends AbstractFileStore {
    private static final String SQL_DEFINE_RESOURCE = "INSERT INTO resources (uri, resource, mime_media_type_id) VALUES (?, ?, SELECT mime_media_type_id FROM mime_media_types WHERE mime_media_type = ?)";
    private static final String SQL_DEFINE_MEDIA_TYPE = "INSERT INTO mime_media_types (mime_media_type) VALUES (?)";
    private static final String SQL_GET_RESOURCE_AND_TYPE = "SELECT r.resource, m.mime_media_type FROM resources r, mime_media_types m WHERE r.uri = ? AND m.mime_media_type_id = r.mime_media_type_id";
    private static final String SQL_UPDATE_MEDIA_TYPE = "UPDATE resources SET mime_media_type_id = (SELECT mime_media_type_id FROM mime_media_types WHERE mime_media_type = ?) WHERE uri = ?";
    private static final String MEDIA_TYPE_UNKNOWN = "application/octet-stream";

    private final Logger log = Logger.getLogger(URIPathFileStore.class);
    private final String contextRoot;
    private final File rootdir;

    private static final Pattern VALID_PATH_ELEMENT = Pattern.compile("[^\\/]*");
    private static final Collection<String> invalidElements = new HashSet<String>();
    static {
	invalidElements.add(".");
	invalidElements.add("..");
    }

    /**
     * Instantiates a FileStore that saves received data in a directory tree that mimics the URI.
     * @param root Root directory for saved files
     * @throws FileNotFoundException if root cannot be created, or exists but is not a directory.
     */
    public URIPathFileStore(final String contextRoot, final File dir) throws FileNotFoundException {
	log.debug("Configuring URIPathFileStore for " + contextRoot + " at " + dir);
	this.contextRoot = contextRoot;
	dir.mkdirs();
	if (!dir.isDirectory()) {
	    throw new FileNotFoundException("not a directory: " + dir);
	}
	this.rootdir = dir;
    }


    /*
     * (non-Javadoc)
     * @see org.nrg.xnat.filestore.FileStore#create(java.sql.Connection, java.net.URI, java.io.File, java.lang.String)
     */
    public void create(final Connection c, final URI uri, final File f, final String mediaType)
    throws SQLException,ForbiddenException {
	register(c, uri, f, null == mediaType ? MEDIA_TYPE_UNKNOWN : mediaType);
    }


    private static String[] decomposePath(final URI uri) throws ForbiddenException {
	final String[] path = uri.toString().split("/");
	for (final String e : path) {
	    if (!VALID_PATH_ELEMENT.matcher(e).matches() || invalidElements.contains(e)) {
		throw new ForbiddenException("invalid path element " + e, uri);
	    }
	}
	return path;
    }


    private File prepare(final URI uri) throws ForbiddenException {
	final String[] path = decomposePath(uri);
	if (0 == path.length) {
	    throw new ForbiddenException("cannot create resource with null path", uri);
	}

	File parent = rootdir;
	for (int i = 0; i < path.length - 1; i++) {
	    final File d = new File(parent, path[i]);
	    if (d.exists()) {
		if (!d.isDirectory()) {
		    throw new ForbiddenException("cannot create (already exists)", d.toURI());
		}
	    } else {
		d.mkdir();
	    }
	    parent = d;
	}
	assert parent.isDirectory();
	return new File(parent, path[path.length - 1]);
    }


    private File prepare(final String uri) throws ForbiddenException {
	try {
	    return prepare(new URI(uri));
	} catch (URISyntaxException e) {
	    throw new ForbiddenException("invalid URI " + uri, null, e);
	}
    }


    /*
     * (non-Javadoc)
     * @see org.nrg.xnat.filestore.FileStore#store(java.sql.Connection, java.net.URI, javax.activation.DataSource)
     */
    public void create(final Connection c, final URI proposed, final DataSource ds)
    throws SQLException,IOException,ForbiddenException {
	final String uri = proposed.toString();
	int i = uri.startsWith(contextRoot) ? contextRoot.length() : 0;
	while (uri.charAt(i) == '/') {
	    i++;
	}
	File outf = prepare(uri.substring(i));

	try {
	    register(c, proposed, outf, null == ds ? MEDIA_TYPE_UNKNOWN : ds.getContentType());
	} catch (SQLException e) {
	    outf.delete();
	    throw e;
	}

	final OutputStream os = new BufferedOutputStream(new FileOutputStream(outf));   
	try {
	    if (null != ds) {
		IOUtils.copy(ds.getInputStream(), os);
	    }
	} finally {
	    os.close();
	}
    }


    private void register(final Connection c, final URI uri, final File resource, final String mediaType)
    throws SQLException,ForbiddenException {
	final String type = defineMediaType(c, mediaType);
	final PreparedStatement s = c.prepareStatement(SQL_DEFINE_RESOURCE);
	try {
	    s.setString(1, uri.toString());
	    s.setString(2, resource.getAbsolutePath());
	    s.setString(3, type);
	    try {
		final int r = s.executeUpdate();
		if (1 != r) {
		    log.error("Unexpected row count defining resource "+ uri + ": " + r);
		}
	    } catch (SQLException e) {
		log.info(e);
		throw new ForbiddenException("resource already defined", uri);
	    }
	} finally {
	    s.close();
	}

	if (log.isDebugEnabled()) {
	    final StringBuilder sb = new StringBuilder("registered resource ");
	    sb.append(uri);
	    sb.append(" -> ");
	    sb.append(resource);
	    sb.append(" (");
	    sb.append(mediaType);
	    sb.append(")");
	    log.debug(sb.toString());
	}
    }


    public boolean update(final Connection c, final URI uri, final DataSource ds)
    throws SQLException,IOException,ForbiddenException {
	final PreparedStatement s = c.prepareStatement(SQL_GET_RESOURCE_AND_TYPE);
	try {
	    s.setString(1, uri.toString());
	    final ResultSet rs = s.executeQuery();
	    try {
		if (!rs.next()) {
		    create(c, uri, ds);
		    return true;
		}
		final String path = rs.getString(1);
		if (null == path) {
		    final String message = "Null local path for exported resource " + uri;
		    log.error(message);
		    throw new FileNotFoundException(message);
		}
		final String contentType = rs.getString(2);
		if (null != ds) {
		    // Save the data
		    final OutputStream os = new BufferedOutputStream(new FileOutputStream(path));   
		    try {
			IOUtils.copy(ds.getInputStream(), os);
		    } finally {
			os.close();
		    }

		    // Update media type, if it has changed
		    final String dsType = ds.getContentType().toLowerCase();
		    if (!dsType.equals(contentType)) {
			defineMediaType(c, dsType);
			final PreparedStatement update = c.prepareStatement(SQL_UPDATE_MEDIA_TYPE);
			try {
			    update.setString(1, dsType);
			    update.setString(2, uri.toString());
			    final int r = update.executeUpdate();
			    if (1 != r) {
				log.error("Unexpected row count updating media type for " + uri + ": " + r);
			    }
			} finally {
			    update.close();
			}
		    }
		}
		return false;
	    } finally {
		rs.close();
	    }
	} finally {
	    s.close();
	}
    }


    private static final String defineMediaType(final Connection c, final String type) throws SQLException {
	final PreparedStatement s = c.prepareStatement(SQL_DEFINE_MEDIA_TYPE);
	final String rtype = null == type ? MEDIA_TYPE_UNKNOWN : type.toLowerCase();
	try {
	    s.setString(1, rtype);
	    try {
		s.executeUpdate();
	    } catch (SQLException ignore) {}	// might already be defined; that's okay.
	    return rtype;
	} finally {
	    s.close();
	}
    }


    protected InputStream getInputStream(final String path) throws FileNotFoundException {
	return new FileInputStream(path);
    }

    protected OutputStream getOutputStream(final String path) throws FileNotFoundException {
	return new BufferedOutputStream(new FileOutputStream(path));
    }


    protected final void deleteResource(final String path) {
	log.debug("removing " + path + " and its parents");
	for (File f = new File(path); null != f && f.delete(); f = f.getParentFile()) {
	    // empty statement; loop removes named file and all empty parent directories
	}
    }

}
