/**
 * Copyright (c) 2008 Washington University
 */
package org.nrg.xnat.tags;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.Statement;
import java.sql.SQLException;
import java.util.Arrays;
import java.util.Collection;
import java.util.ArrayList;

import org.apache.log4j.Logger;
import org.apache.commons.lang.StringEscapeUtils;


/**
 * @author Kevin A. Archie <karchie@npg.wustl.edu>
 *
 */
public final class TagOps {
    private static final String SQL_MATCH_LABEL = "SELECT tag_label FROM tag_labels WHERE tag_label_lc = LOWER(?)";
    private static final String SQL_ALL_LABELS = "SELECT tag_label FROM tag_labels";
    private static final String SQL_ADD_LABEL = "INSERT INTO tag_labels (tag_label, tag_label_lc) VALUES ( ?, LOWER(?) )";
    private static final String SQL_DELETE_TAGS =
	"DELETE FROM resource_tags WHERE resource_tags.tag_label_id = (SELECT tag_label_id FROM tag_labels WHERE tag_label_lc = LOWER(?))";
    private static final String SQL_DELETE_LABEL = "DELETE FROM tag_labels WHERE tag_label_lc = LOWER(?)";


    /**
     * Defines the given tag labels in the database.
     * @param ds Connection to database
     * @param labels TagLabels to define
     * @throws SQLException
     */
    public static void add(final Connection c, final Iterable<TagLabel> labels) throws SQLException {
	final PreparedStatement doAdd = c.prepareStatement(SQL_ADD_LABEL);
	try {
	    for (final TagLabel tagLabel : labels) {
		final String sl = tagLabel.toString();
		assert sl.equals(StringEscapeUtils.escapeSql(sl));
		doAdd.setString(1, sl);	// first is for retained value
		doAdd.setString(2, sl);	// second is for normalized value
		try {
		    if (1 == doAdd.executeUpdate()) {
			Logger.getLogger(TagOps.class).debug("tag label " + sl + " created");
		    } else {
			Logger.getLogger(TagOps.class).debug("tag label " + sl + " not created");
		    }
		} catch (SQLException e) {
		    Logger.getLogger(TagOps.class).debug("tag label " + sl + " not created", e);
		}
	    }
	} finally {
	    try { doAdd.close(); } catch (SQLException ignore) {}
	}
    }


    /**
     * Defines the given tag labels in the database.
     * @param c Connection to database
     * @param labels TagLabels to define
     * @throws SQLException
     */
    public static void add(final Connection c, final TagLabel...labels) throws SQLException {
	add(c, Arrays.asList(labels));
    }


    /**
     * Deletes the named tag labels and all associated tags from the database
     * @param c Connection to database
     * @param labels tag labels to be deleted
     * @return tag labels that were succesfully deleted
     * @throws SQLException
     */
    public static Collection<TagLabel> delete(final Connection c, final Iterable<TagLabel> labels) throws SQLException {
	final PreparedStatement deleteTags = c.prepareStatement(SQL_DELETE_TAGS);
	try {
	    final PreparedStatement deleteLabel = c.prepareStatement(SQL_DELETE_LABEL);
	    try {
		final Collection<TagLabel> deleted = new ArrayList<TagLabel>();
		for (final TagLabel tagLabel : labels) {
		    final String sl = tagLabel.toString();
		    assert sl.equals(StringEscapeUtils.escapeSql(sl));

		    // Remove matching tags from resource metadata
		    deleteTags.setString(1, sl);
		    deleteTags.executeUpdate();

		    // Remove matching label from label definitions
		    deleteLabel.setString(1, sl);
		    int rows = deleteLabel.executeUpdate();
		    switch (rows) {
		    case 0:
			break;

		    case 1:
			deleted.add(tagLabel);
			break;

		    default:
			Logger.getLogger(TagOps.class).error("Strange row count for label DELETE: " + rows);
		    deleted.add(tagLabel);
		    break;
		    }
		}
		return deleted;
	    } finally {
		try { deleteLabel.close(); } catch (SQLException ignore) {}
	    }
	} finally {
	    try { deleteTags.close(); } catch (SQLException ignore) {}
	}
    }


    /**
     * Deletes the named tag labels from the database.
     * @param c Connection to database
     * @param labels tag labels to be deleted
     * @return tag labels that were succesfully deleted
     * @throws SQLException
     */
    public static Collection<TagLabel> delete(final Connection c, final TagLabel...labels)
    throws SQLException {
	return delete(c, Arrays.asList(labels));
    }


    /**
     * Retrieves all tag labels defined in this database
     * @param c Connection to database
     * @return tag labels defined in the database
     * @throws SQLException
     */
    public static Collection<TagLabel> getAll(final Connection c) throws SQLException {
	final Statement s = c.createStatement();
	try {
	    final ResultSet rs = s.executeQuery(SQL_ALL_LABELS);
	    try {
		final Collection<TagLabel> labels = new ArrayList<TagLabel>();
		while (rs.next()) {
		    try {
			labels.add(new TagLabel(rs.getString(1)));
		    } catch (InvalidTagLabelException e) {
			throw new RuntimeException("Database inconsistency: found invalid tag label in database", e);
		    }
		}
		return labels;
	    } finally {
		try { rs.close(); } catch (SQLException ignore) {}
	    }
	} finally {
	    try { s.close(); } catch (SQLException ignore) {}
	}
    }


    /**
     * Retrieves all tag labels from the given collection that are defined in the database.
     * @param c Connection to database
     * @param labels tag labels to be checked
     * @return all requested tag labels found in the database, with canonical capitalization
     * @throws SQLException
     */
    public static Collection<TagLabel> getMatching(final Connection c, final Iterable<TagLabel> labels)
    throws SQLException {
	final PreparedStatement doSearch = c.prepareStatement(SQL_MATCH_LABEL);
	try {
	    final Collection<TagLabel> found = new ArrayList<TagLabel>();
	    for (final TagLabel tagLabel : labels) {
		final String sl = tagLabel.toString();
		assert sl.equals(StringEscapeUtils.escapeSql(sl));
		doSearch.setString(1, sl);
		final ResultSet rs = doSearch.executeQuery();
		try {
		    if (rs.next()) {
			try {
			    final TagLabel canonical = new TagLabel(rs.getString(1));
			    found.add(canonical);
			} catch (InvalidTagLabelException e) {
			    Logger.getLogger(TagOps.class).error("error generating canonical label", e);
			    found.add(tagLabel);	// return something despite the weirdness
			}
			assert !rs.next();
		    }
		} finally {
		    try { rs.close(); } catch (SQLException ignore) {}
		}
	    }
	    return found;
	} finally {
	    try { doSearch.close(); } catch (SQLException ignore) {}
	}
    }


    /**
     * Retrieves all tag labels from the given array that are defined in the database.
     * @param c Connection to database
     * @param labels tag labels to be checked
     * @return all requested tag labels found in the database, with canonical capitalization
     * @throws SQLException
     */
    public static Collection<TagLabel> getMatching(final Connection c, final TagLabel...labels)
    throws SQLException {
	return getMatching(c, Arrays.asList(labels));
    }
}
